/*
//
// BEGIN SONGBIRD GPL
//
// This file is part of the Songbird web player.
//
// Copyright(c) 2005-2008 POTI, Inc.
// http://songbirdnest.com
//
// This file may be licensed under the terms of of the
// GNU General Public License Version 2 (the "GPL").
//
// Software distributed under the License is distributed
// on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied. See the GPL for the specific language
// governing rights and limitations.
//
// You should have received a copy of the GPL along with this
// program. If not, go to http://www.gnu.org/licenses/gpl.html
// or write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
// END SONGBIRD GPL
//
*/

/**
 * \file  DaapImporterJob.jsm
 * \brief Daap Job for importing remote tracks on a local library
 */

//------------------------------------------------------------------------------
//
// Exported object
//
//------------------------------------------------------------------------------

EXPORTED_SYMBOLS = [ "DaapImporterJob" ];

//------------------------------------------------------------------------------
//
// DaapImporterJob definition
//
//------------------------------------------------------------------------------



if (typeof Cc == 'undefined') var Cc = Components.classes;
if (typeof Ci == 'undefined') var Ci = Components.interfaces;
if (typeof Cu == "undefined") var Cu = Components.utils;
if (typeof Cr == 'undefined') var Cr = Components.results;

/* global variables */
var ioService = null;

/* Make sure we have the javascript modules we're going to use */
if (typeof(XPCOMUtils) == "undefined") {
  Cu.import("resource://gre/modules/XPCOMUtils.jsm");
  if (!XPCOMUtils)
    throw new Error("Import of XPCOMUtils module failed!");
}
if (typeof(SBJobUtils) == "undefined") {
  Cu.import("resource://app/jsmodules/SBJobUtils.jsm");
  if (!XPCOMUtils)
    throw new Error("Import of SBJobUtils module failed!");
}

/* based on sbITunesImporterJob */


function DaapImporterJob() {
  // Call super-class constructor.
  SBJobUtils.JobBase.call(this);
}


DaapImporterJob.prototype = {
  // Inherit prototype.
  __proto__: SBJobUtils.JobBase.prototype,

  //   _updateProgressDelay     Time in milliseconds to delay job progress
  //                            updates.
  _updateProgressDelay: 1000,

  //    _nice                   more greater, more nicer
  _nice:  20,


  //   _updateProgressTimer     Timer used to delay job progress updates to
  //                            limit update rate.
  //   _abort                   if the current job must be aborted or not
  _updateProgressTimer: null,
  _abort: false,


  //----------------------------------------------------------------------------
  //
  // Public importer job services.
  //
  //----------------------------------------------------------------------------

  /**
   * Set the job status to the sbIJobProgress status value specified by aStatus.
   *
   * \param aStatus             Job status.
   */
  setStatus: function DaapImporterJob_setStatus(aStatus) {
    this._status = aStatus;
    // this._updateProgress();
    this._updateProgressDelayed(); //update now!
  },

  /**
   * Set the job status text to the value specified by aStatusText.
   *
   * \param aStatusText         Job status text.
   */
  setStatusText: function DaapImporterJob_setStatusText(aStatusText) {
    this._statusText = aStatusText;
    this._updateProgress();
  },

  /**
   * Set the job title text to the value specified by aTitleText.
   *
   * \param aTitleText          Job title text.
   */
  setTitleText: function DaapImporterJob_setTitleText(aTitleText) {
    this._titleText = aTitleText;
    this._updateProgress();
  },

  /**
   * Set the job work units completed progress value to the value specified by
   * aProgress.
   *
   * \param aProgress           Job work units completed progress.
   */
  setProgress: function DaapImporterJob_setProgress(aProgress) {
    this._progress = aProgress;
    this._updateProgress();
  },

  /**
   * Set the job total work units to complete value to the value specified by
   * aTotal.
   *
   * \param aTotal              Total job work units to complete.
   */
  setTotal: function DaapImporterJob_setTotal(aTotal) {
    this._total = aTotal;
    this._updateProgress();
  },

  /**
   * Set if this job can/cannot be canceled.
   *
   * \param canCancel   true/false
   */
  setCanCancel: function DaapImporterJob_setCanCancel(canCancel) {
    if (canCancel) {
      this._canCancel = true;
    }
    else {
      this._canCancel = false;
    }
  },

  /**
   * job aborted getter
   */
  get abort() {
    return this._abort;
  },

  //----------------------------------------------------------------------------
  //
  // Importer job sbIJobCancelable services.
  //
  //----------------------------------------------------------------------------

  /**
   * \brief Attempt to cancel the job
   * Throws NS_ERROR_FAILURE if canceling fails
   */
  cancel: function DaapImporterJob_cancel() {
    this._abort = true;
  },


  //----------------------------------------------------------------------------
  //
  // Internal importer job services.
  //
  //----------------------------------------------------------------------------

  /**
   * Update the job progress.  This function delays a bit before updating to
   * limit the update rate.
   */
  _updateProgress: function DaapImporterJob__updateProgress() {
    // Do nothing if update progress timer is already running.
    if (this._updateProgressTimer)
      return;

    // Delay before updating progress to limit progress update rate.
    var _this = this;
    var func = function() { _this._updateProgressDelayed(); };
    this._updateProgressTimer =
           Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    this._updateProgressTimer.initWithCallback(func,
                                               this._updateProgressDelay,
                                               Ci.nsITimer.TYPE_ONE_SHOT);
  },

  _updateProgressDelayed:
    function DaapImporterJob__updateProgressDelayed() {
    this.notifyJobProgressListeners();
    this._beNice();

    // keep this at the end
    this._updateProgressTimer = null;
  },


  _beNice: function DaapImporterJob__beNice ()
  {
    /*  be a little "polite" an let others work */
    let thread = Components.classes["@mozilla.org/thread-manager;1"]
                           .getService().currentThread;
    for (let i=0; i<this._nice && thread.hasPendingEvents(); i++) {
        thread.processNextEvent(false);
    }

  },

  /* start:
   *  Start the import job.
   *
   *  \param  aClient     an initialized daap client
   *  \param  aMediaList  the medialist to import
   *  \param  aCount      expected number of items
   *
   **/
  start: function DaapImporterJob_start (aClient, aMediaList, aCount)
  {
    /* preconditions */
    if (!(aClient instanceof Ci.sbIDaapClient)
        || !(aMediaList instanceof Ci.sbIMediaList) 
        || !(typeof(aCount) == "number" && aCount >=0) ) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    /* import modules */
    if (typeof(SBProperties) == "undefined") {
      Components.utils.import("resource://app/jsmodules/sbProperties.jsm");
    }
    if (typeof(DProperty) == "undefined") {
      Components.utils.import("resource://daapclient/DaapProperties.jsm");
    }

    /* create the property array */
    let properties = Cc["@mozilla.org/hash-property-bag;1"]
                        .createInstance(Ci.nsIWritablePropertyBag2);

    /* setup job data */
    this.setTotal(aCount);
    this.setStatus(Ci.sbIJobProgress.STATUS_RUNNING);
    this.setCanCancel(true);

    let id = aMediaList.getProperty(DProperty.id);
    if (!id) {
      /* abort */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* start the clock for debbuging purposes */
    if (DaapDebug) this._start = Date.now();

    if (aMediaList instanceof Ci.sbILibrary) {
      var listener = new DaapBatchListener(this);
      if (!listener) {
        throw Components.results.NS_ERROR_OUT_OF_MEMORY;
      }


      /* request all available metadata for library items */
      properties.setPropertyAsBool(SBProperties.trackName, true);
      properties.setPropertyAsBool(SBProperties.artistName, true);
      properties.setPropertyAsBool(SBProperties.albumArtistName, true);
      properties.setPropertyAsBool(SBProperties.albumName, true);
      properties.setPropertyAsBool(SBProperties.duration, true);
      properties.setPropertyAsBool(SBProperties.contentLength, true);
      properties.setPropertyAsBool(SBProperties.genre, true);
      properties.setPropertyAsBool(SBProperties.bpm, true);
      properties.setPropertyAsBool(SBProperties.bitRate, true);
      properties.setPropertyAsBool(SBProperties.comment, true);
      properties.setPropertyAsBool(SBProperties.isPartOfCompilation, true);
      properties.setPropertyAsBool(SBProperties.created, true);
      properties.setPropertyAsBool(SBProperties.updated, true);
      properties.setPropertyAsBool(SBProperties.totalDiscs, true);
      properties.setPropertyAsBool(SBProperties.discNumber, true);
//    properties.setPropertyAsBool(SBProperties., true);
//    properties.setPropertyAsBool(SBProperties., true);
//    properties.setPropertyAsBool(SBProperties., true);
//    properties.setPropertyAsBool(SBProperties., true);
      properties.setPropertyAsBool(SBProperties.sampleRate, true);
//    properties.setPropertyAsBool(SBProperties., true);
//    properties.setPropertyAsBool(SBProperties., true);
      properties.setPropertyAsBool(SBProperties.totalTracks, true);
      properties.setPropertyAsBool(SBProperties.trackNumber, true);
      properties.setPropertyAsBool(SBProperties.rating, true);
      properties.setPropertyAsBool(SBProperties.year, true);
//    properties.setPropertyAsBool(SBProperties., true);
      properties.setPropertyAsBool(SBProperties.originURL, true);
      properties.setPropertyAsBool(SBProperties.composerName, true);

         
      // XXX: TODO: make a config option for "pace" and "politeness" ;-)
      aClient.batchCreateMediaItems(id, properties, aMediaList, listener, 500);
    }
    else {
      let databaseId = aMediaList.library.getProperty(DProperty.id);
      if (!databaseId) {
        /* abort */
        throw Components.results.NS_ERROR_FAILURE;
      } 

      var listener = new DaapAsyncListener(this, aMediaList);
      if (!listener) {
        throw Components.results.NS_ERROR_OUT_OF_MEMORY;
      }

      /* don't request metadata of playlist items for speeding up */

      aClient.getPlaylistItemsAsync(listener, databaseId, id, properties);
    }

    // thanks Mook_sb!!
    let sip = Cc["@mozilla.org/supports-interface-pointer;1"]
               .createInstance(Ci.nsISupportsInterfacePointer);
    sip.data = this;
    SBJobUtils.showProgressDialog(sip.data, null, 20);

  }, //start


  onComplete: function DaapBatchListener_onComplete (aMediaItems, aResult)
  {
    if (aResult != Cr.NS_OK) {
      /* not succeded */
      this.setStatus(Ci.sbIJobProgress.STATUS_FAILED);

      /* log the status for debugging  */
      if (aResult == Cr.NS_ERROR_ABORT) {
        /* aborted */
        if (DaapDebug) DaapDebug("import aborted.", this._start);
      }
      else {
        /* error */
        if (DaapDebug) DaapDebug("import error.", this._start);
      }

      /* nothing more to do here */
      return;
    }

    /* 
     * importing succeded 
     *
     **/

    /* finish the job */
    this.setStatus(Ci.sbIJobProgress.STATUS_SUCCEEDED);

    /* display debug data */
    if (DaapDebug) DaapDebug("import finished.", this._start);

    return;
  }, 

};


//XXXDaniel fix getErrorMessages to return a nsIStringEnumerator as
// sbIJobProgress say TODO: file a bug

DaapImporterJob.prototype.getErrorMessages = function Job_getErrorMessages_fix () {
  if (typeof(ArrayConverter) == "undefined") {
    Cu.import("resource://app/jsmodules/ArrayConverter.jsm");
  }
  return (this._jobProgressDelegate) ? 
          this._jobProgressDelegate.getErrorMessages() :
          ArrayConverter.stringEnumerator(this._errorMessages);
};




//------------------------------------------------------------------------------
//
// DaapBatchListener definition (internal)
//
//------------------------------------------------------------------------------

function DaapBatchListener (aJob)
{
  if (! (aJob instanceof DaapImporterJob)) {
    throw Components.results.NS_ERROR_INVALID_ARG;
  }

  this._job = aJob;
}

DaapBatchListener.prototype = {
  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIBatchCreateMediaItemsListener]),

  /* private data ---------------------------------------------------------*/
  _job: null,                  

  /* sbIBatchCreateMediaItemsListener -------------------------------------*/
  onProgress: function DaapBatchListener_onProgress (aCount)
  {
    if (this._job.abort) {
      throw Cr.NS_ERROR_ABORT;
    }

    /* update progress */
	  this._job.setProgress(aCount);
  },

  onComplete: function DaapBatchListener_onComplete (aMediaItems, aResult)
  {
    /* notify the job we're done */
    this._job.onComplete(aMediaItems, aResult);
  }, 

};



//------------------------------------------------------------------------------
//
// DaapAsyncListener definition (internal)
//
//------------------------------------------------------------------------------

function DaapAsyncListener (aJob, aMediaList)
{
  if (!(aJob instanceof DaapImporterJob)
      || !(aMediaList instanceof Ci.sbIMediaList) ) {
    throw Components.results.NS_ERROR_INVALID_ARG;
  }

  if (!(ioService instanceof(Ci.nsIIOService))) {
    ioService = Components.classes["@mozilla.org/network/io-service;1"]
                          .getService(Ci.nsIIOService);
    if (!ioService) {
      throw Components.results.NS_ERROR_FAILURE;
    }
  }

  this._job = aJob;
  this._mediaList = aMediaList;
  this._mediaItems = null;

}

DaapAsyncListener.prototype = {

  /* private data ---------------------------------------------------*/
  _job: null,
  _mediaList: null,
  _mediaItems: null,

  /* nsISupports ----------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapItemsListener]),

  /* sbIDaapItemsListener -------------------------------------------*/
  onEnumerationBegin: function DaapAsyncListener_onEnumerationBegin () 
  {
    this._mediaItems = Cc["@mozilla.org/array;1"]
                          .createInstance(Ci.nsIMutableArray);
    if (!this._mediaItems) {
      throw Components.results.NS_ERROR_OUT_OF_MEMORY;
    }

  },


  onEnumeratedItem: function DaapAsyncListener_onEnumeratedItem (aItemURL, aProperties) 
  {
    //FIXME dump("CREATE: "+aItemURL+"\n");

    if (this._job.abort) {
      throw Cr.NS_ERROR_ABORT;
    }

   
    let mediaItem = null;
    if (aItemURL) {
      /* create a new media item */
      let itemURI = ioService.newURI(aItemURL, null, null);
      mediaItem = this._mediaList.library.createMediaItem(itemURI, aProperties, false);
    }
    else {
      //FIXME: rhythmbox FIX
      try {
        aProperties.appendProperty(SBProperties.isList, "0");
        let mediaItems = this._mediaList.library.getItemsByProperties(aProperties);
        if (mediaItems && mediaItems.length) {
          mediaItem = mediaItems.queryElementAt(0, Ci.sbIMediaItem);
        }
      }
      catch (e) {}
    }

    if (!mediaItem) {
      /* abort */
      return;
    }

    this._mediaList.add(mediaItem);
    this._mediaItems.appendElement(mediaItem, false);

    /* update progress */
	  this._job.setProgress(this._mediaItems.length);

    //dump("CREATE I'm nice : "+aItemURL+"\n");
    /*  be a little "polite" an let others work */
    //FIXME!!
    /* 
    let thread = Components.classes["@mozilla.org/thread-manager;1"]
                           .getService().currentThread;
    for (let i=0; i<15 && thread.hasPendingEvents(); i++) {
        thread.processNextEvent(false);
    }
    */
    
    //dump("CREATE FINISHED!: "+aItemURL+"\n");
  
  }, //onEnumeratedItem


  onEnumerationEnd: function DaapAsyncListener_onEnumerationEnd (aCount)
  {
    /* update job status */
    if (this._mediaItems.length == this._job.total) {
      this._job.onComplete(this._mediaItems, Components.results.NS_OK);
    }
    else {
      this._job.onComplete(this._mediaItems, Components.results.NS_ERROR_FAILURE);
    }
  }, //onEnumerationEnd

};



/**
 * Display debug messages and stats
 *
 **/

if (typeof(DaapDebug) == 'undefined') {

  function DaapDebug(aMessage, aStart) {

    if (aStart) {
      var elapsed = Date.now() - aStart;
      var time = (elapsed / 1000);
      aMessage += " (time: " + time + ")";
    }

    var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
                                 .getService(Components.interfaces.nsIConsoleService);
    consoleService.logStringMessage("[Daap Client] " + aMessage);
  }  
}


 

